/*
*	extended Karp-Rabin algorithm
*
*	handbook, p.31
*
*	search algorithm of O(n)
*	where n is the size of text
*
*	very smart algorithm!
*	applicable to the search of multi-patterns of same length
*
*	I have extended the algorithm to multi-patterns of variable length
*	and applied to string translation problem
*	-Kenar-
*/

#include <u.h>
#include <libc.h>
#include <ctype.h>
#include <bio.h>
#include "trans.h"

#define DBG if(0)
#define	DBG0 if(0)

Biobuf bout;

typedef struct Hkeys Hkeys;
struct Hkeys{
	char *key;
	void *value;
	Hkeys *next;
};
Hkeys **hkeys0, **hkeys1;

/*	nhkey should be prime number */
int nhkeys;

int pow2[256];

int hashlen;
char *sep;

/*	a simple linear hash */
int
mhash(char *s, int n)
{
	int h = 0;
	while(n-- > 0) h = (h << 1)%nhkeys + *s++;
	h %= nhkeys;
	return h<0?h+nhkeys:h;
}

int
rehash(char *s, int m, int h)
{
	/*
	*	m is strlen of the pattern
	*
	*	NB: pow() returns double	*/
	h = (((h - *(s-1) * pow2[m-1]) << 1) + *(s+m-1)) % nhkeys;
	return h<0?h+nhkeys:h;
}

int
hash(char *s)
{
	return mhash(s, strlen(s));
}


void
Hmput(Hkeys **hdb, char *key, int m, void *value)
{
	int h = mhash(key,m);
	Hkeys *p = new(Hkeys);
	p->key = key;
	p->value = value;
	p->next = hdb[h];
	hdb[h] = p;
}

void
Hvput(Hkeys **hdb, char *key, int m)
{
	int h = mhash(key,m);
	Hkeys *p = new(Hkeys);
/*
*	p->key is unused
*/
	p->key = key;
	p->value = key;
	p->next = hdb[h];
	hdb[h] = p;
}


void
Hput(Hkeys **hdb, char *key, void *value)
{
	int h = hash(key);
	Hkeys *p = new(Hkeys);
	p->key = key;
	p->value = value;
	p->next = hdb[h];
	hdb[h] = p;
}

void *
Hmget(Hkeys **hdb, char *key, int m)
{
	Hkeys *p;
	for(p = hdb[mhash(key,m)];p;p = p->next)
		if(strcmp(p->key, key)==0)
			return p->value;
	return nil;
}

void *
Hget(Hkeys **hdb, char *key)
{
	Hkeys *p;
	for(p = hdb[hash(key)];p;p = p->next)
		if(strcmp(p->key, key)==0)
			return p->value;
	return nil;
}

void *
Hpull(Hkeys **hdb, int h, char *key)
{
	Hkeys *p;
	for(p = hdb[h];p;p = p->next)
		if(strcmp(p->key, key)==0)
			return p->value;
	return nil;
}

/*	extract first match value
*	values be sorted reverse dictionary order
*	in order the return value has any meaning	*/
char *
Hvpull(Hkeys **hdb, int h, char *text)
{
	Hkeys *p;
	char *q;
	for(p = hdb[h];p;p = p->next){
		q = p->value;
		/* it is curious, strncmp is faster than memcmp */
		//if(memcmp(q, text, strlen(q))==0)
		if(strncmp(q, text, strlen(q))==0)
			return q;
	}
	return nil;
}



/*
*	tabread: read translation data from rule file
*	lines beginning with "#" are comment
*	if the first line is comment and string "unicode" or "hex" is in it,
*	these are keywords for format of the file 	*/
void
tabread(char *path)
{
	int n,i,min;
	Trtab trtab;
	
	n = readtable0(&trtab,path,1);
	ttab = trtab.ttab;
	min = trtab.min;

	nhkeys = upprime(n);

	hkeys0 = (Hkeys**) calloc(nhkeys, sizeof(Hkeys*));
	hkeys1 = (Hkeys**) calloc(nhkeys, sizeof(Hkeys*));

	for(i = 0; i < n ; i++)
		Hput(hkeys0, ttab[i].frstr, ttab[i].tostr);

	hashlen = min;

	/*	hkeys1 is a database that stores trancated frstr in addition to full frstr.
	*	this database exist so that we can extract
	*	full frstr from trancated data.
	*	ttab[i].frstr must be sorted in dictionary order
	*	so that we can easily extract longest frstr that matches
	*	string in the text
	*	note that Hvput() stores data next to it's root	*/
	for(i = 0; i < n ; i++)
		Hvput(hkeys1, ttab[i].frstr, hashlen);
	nrule = n;
	DBG0 print("done\n");
}


/*	advanced: O(n)	*/
#define OUTPUT(s,n) Bwrite(&bout,s,n)

void
trans(char *text)
/*	
*	source: source text
*	ttab: table of rule
*	n: number of rules	*/
{
	int h,m;
	char *s,*t,*q;
	char *end = text + strlen(text) - hashlen;
	m = hashlen; /* m is const */
	h = -1;
	for(t = text; t <= end; t++){
		DBG print("0> %s\n", t);
		if(h == -1)
			h = mhash(t,m);
		else{
			h = rehash(t,m,h);
			DBG print("%d %d\n",h,mhash(t,m));
		}
		q = Hvpull(hkeys1,h,t);
		if(q){
				s = Hget(hkeys0,q);
				DBG print("2> %s %s\n",q,s);
				OUTPUT(s, strlen(s));
				t += strlen(q) - 1; /* NB: next t++ */
				h = -1;
				continue;
		}
		OUTPUT(t, 1);
	}
	OUTPUT(t, strlen(t));
}


/*
*	translate() shows the usage example of etrans()
*	etrans() is designed for string to string conversion
*	these string must be utf-8.
*/
void
translate(int fd)
{
	Biobuf bin;
	char *source;
	if(Binit(&bin, fd, OREAD) < 0)
		sysfatal("input not open: %r");
	while((source = Brdstr(&bin,'\n',1))){ /* assign = */
		trans(source);
		Bwrite(&bout,"\n",1);
	}
	Bflush(&bout);
}


void
usage(void)
{
	fprint(2,
	"usage: eKR_trans [-F sep] rulefile [file ....]\n"
	);
	exits("usage");
}

void
main(int argc, char *argv[])
{
	int i, fd;
	ARGBEGIN{
	case 'F': sep = EARGF(usage());
		break;
	default: usage();
	}ARGEND

	if(*argv == nil)
		usage();

	Binit(&bout, 1, OWRITE);

	tabread(*argv++);

	pow2[0] = 1;
	for(i = 1; i < 256; i++)
		pow2[i] = pow2[i-1]*2 % nhkeys;

	if(*argv == nil){
		translate(0);
		exits(nil);
	}

	while(*argv){
		fd = open(*argv, OREAD);
		translate(fd);
		argv++;
	}
	exits(nil);
}
